//-----------------------------------------------------------------------------
//
//  DyPERS.c++
//
//  Dynamic Enhanced Personal Reality System
//  Based on the record and associate paradigm.
//
//  Tony Jebara (Copyright (C) 1997)
//
//-----------------------------------------------------------------------------


// connect to vm15/bernt/C++/Src/DyPERS/    ObjDypers param
// connect to ~bernt/DemoMIT/DyPERS/  then type ./RUN

#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <termios.h>
#include <unistd.h>
#include <rpc/rpc.h>


#include "dypersClient.h"

#include "Video_iris64.h++"
#include "EasyWindow.h++"
#include "Audio.h++"

extern "C" {
#include <gl/gl.h>
#include <dmedia/vl_vino.h>
}
// RPC stuff
// Server includes


// Configuration
char config[256];

// Video Interface
int  vinput     = -2;
int  zoom       =  1;
int  channel    =  2;
int  display_on =  1;
int  sscale     =  3;
int  keyboard   = -1;
int  ss         =  4;

#define MAXPICS  16384
#define MAXCLIPS  1024
#define MAXLABELS 1024
#define MAXAUDIO 2000000

#define tmax(a,b) ((a)>(b)) ? (a):(b)

// Record and Associate Counters
int  recClip   = 0;
int  recPic    = 0;
int  playFrame = -1;
int  playClip  = -1;

// Video Sequences and Control
int     vw, vh;
int     w, h;
ulong  *current;
ulong  *blitter;
ulong  *blacker;
char   *vidpics[MAXPICS];
int     clipstart[MAXCLIPS];
int     clipstop[MAXCLIPS];
int     clippause[MAXCLIPS];
long    pictime[MAXPICS];
struct  timeval time1;
long    starttime;
int     objectlabel=-1;
int     winmoved=-1;

// Audio Sequences and Control
float  *audioclips[MAXCLIPS];
float  *audioclip;
float  *audioptr;
int     audiolen[MAXCLIPS];
int     totalframes = 0;
int     audioFrame;
// Camera
SGICam *v;

// Window
long    winmovie;

// Audio
AudioIn mic;

// RPC variables
dypersClient        cldypers;
dypersRecResult     recRes;
char*               server = NULL;

void
associate()
{
  if (server!=NULL) cldypers.learnObject(recClip);
}

void
garbage()
{
  if (server!=NULL) cldypers.learnObject(-1);
}

void
record()
{
  char  *vidpic;
  int    y;
  ulong *z1;
  char  *z2, *z3;
  ulong  val;

  vidpics[recPic] = (char *) malloc(w*h*sizeof(char));
  vidpic          = vidpics[recPic];
  
  gettimeofday(&time1);
  pictime[recPic] = time1.tv_sec*1000000+time1.tv_usec;

  for (y=0; y<h; y++)
    {
      z1 = &current[(y*sscale)*vw];
      z2 = &vidpic[y*w];
      z3 = &vidpic[y*w+w];
      for (; z2<z3; z2++)
	{
	  val   = *(z1);
	  *(z2) = ( ((val & 0xC00000) >> 16) +
		    ((val & 0x00E000) >> 10) +
		    ((val & 0x0000E0) >> 5) );
	  z1 += sscale;
	}
    }

  recPic++;
  clipstop[recClip] = recPic;
}

int
render(int n)
{
  char *vidpic;
  int   i, val;
  long  activewindow;



  if ((n>=0) && (n<MAXPICS))
    if (vidpics[n]!=NULL)
      {

	vidpic = vidpics[n];
	for (i=0; i<(w*h); i++)
	  {
	    val = vidpic[i];
	    blitter[i] = 0xFF201010 + ( ((val & 0xC0) << 16) +
					((val & 0x38) << 10) +
					((val & 0x07) << 5) );
	  }
	activewindow = winget();
	winset(winmovie);
	if (winmoved != 1)
	  {
	    winmove(10, 640);
	    winmoved = 1;
	  }
	lrectwrite(0,0, w-1, h-1, (ulong *) blitter);
	winset(activewindow);
	return(1);
      }
  
  activewindow = winget();
  winset(winmovie);
  if (winmoved != 2)
    {
      winmove(0, 0);
      winmoved = 2;
    }
  lrectwrite(0,0, w-1, h-1, (ulong *) blacker);
  winset(activewindow);
  return(-1);
}

void
playback(int n)
{
  if ((playClip==n) || (n < 0))
    {
      if ((playClip>=0) && (playClip<MAXCLIPS))
	{
	  playFrame++;
	  if ((playFrame>=0) &&
	      (playFrame<clipstop[playClip]) &&
	      (playFrame>=clipstart[playClip]))
	    {
	      gettimeofday(&time1);
	      if (playFrame==clipstart[playClip])
		starttime = time1.tv_sec*1000000+time1.tv_usec;
	      else
		usleep(tmax((pictime[playFrame]-pictime[clipstart[playClip]])
			    -(time1.tv_sec*1000000+time1.tv_usec - starttime),0));
	      render(playFrame);
	      audioFrame = mic.Write(audiolen[playClip],audioclips[playClip], audioFrame);
	    }
	  else
	    {
	      playClip = -1;
	    }
	}
      if (n<0)
	{
	  render(-10);
	  playClip = -1;
	}
    }
  else
    {
      playClip   = n;
      playFrame  = clipstart[playClip];
      gettimeofday(&time1);
      if (playFrame==clipstart[playClip])
	starttime = time1.tv_sec*1000000+time1.tv_usec;
      audioFrame = 0;
      audioFrame = mic.Write(audiolen[playClip],audioclips[playClip], audioFrame);
      // printf("New audioFrame=%d\n",audioFrame);
      render(playFrame);
    }
}

void
parse_args (int argc, char **argv)
{
  int i;

  // Set some defaults
  strcpy(config,"");


  // Set the command line arguments
  for(i=1; i<argc; i++)
    {
      if (!strcmp(argv[i], "-input"))
	vinput = atoi(argv[++i]);
      else if (!strcmp(argv[i], "-zoom"))
	zoom = atoi(argv[++i]);
      else if (!strcmp(argv[i], "-scale"))
	sscale = atoi(argv[++i]);
      else if (!strcmp(argv[i], "-server"))
	server = argv[++i];
      else if (!strcmp(argv[i], "-keyboard"))
	keyboard = 1;
      else if (!strcmp(argv[i], "-channel"))
	channel = atoi(argv[++i]);
      else if (!strcmp(argv[i], "-config"))
	strcpy(config,argv[++i]);
      else 
	{
	  fprintf(stderr,"Usage: %s -input <int> -channel <int> -zoom <int> -scale <int> -keyboard -config <fname> -server <servername>\n",argv[0]);
	  fprintf(stderr,"\nDyPERS !!! \n \n");
	  fprintf(stderr,"Dynamic Personal Enhanced Reality System. \n\n");
	  fprintf(stderr,"The paradigm here is Record and Associate.\n");
	  fprintf(stderr,"Push the right mouse button to record video and audio.\n");
	  fprintf(stderr,"Push the left mouse button to associate that recording with\n");
	  fprintf(stderr,"a picture in the objective.\n");
	  fprintf(stderr,"Push 'q' to quit and save configuration.\n\n");
	  fprintf(stderr,"Tony Jebara \n");
	  fprintf(stderr,"Nuria Oliver \n");
	  fprintf(stderr,"Berndt Schiele \n");
	  exit (2);
	}
    }
  fprintf(stderr,"\nDyPERS - Dynamic Personal Enhanced Reality System. \n");
  fprintf(stderr,"Tony Jebara, Nuria Oliver and Berndt Schiele \n");
}


// For Keyboard Interface
struct termios TermDescr;
struct termios OrigTerm;
int TermMask;
int Key;
int ch;


void
main(int argc, char **argv) 
{
  int     i;

  // Parse Arguments
  parse_args(argc,argv);

  // If server, then connect to it
  if (server!=NULL) cldypers.connect(server);

  // Video and Image Geometry and Declarations
  v       = new SGICam(channel,vinput,zoom);
  vw      = v->width();
  vh      = v->height();
  w       = vw/sscale;
  h       = vh/sscale;
  blitter = (ulong *) malloc(w*h*sizeof(ulong));
  blacker = (ulong *) malloc(w*h*sizeof(ulong));
  for (i=0; i<(w*h); i++)
    {
      blitter[i] = 0;
      blacker[i] = 0;
    }
  prefsize(w,h);
  winmovie = winopen("Movie Window");
  RGBmode();
  RGBcolor(0,100,200);
  pixmode(PM_TTOB, 1);
  gconfig();

  // Initialize the Clips
  for (i=0; i<MAXPICS; i++)   vidpics[i]    = NULL;
  for (i=0; i<MAXCLIPS; i++)  clipstart[i]  = -1;
  for (i=0; i<MAXCLIPS; i++)  clipstop[i]   = -2;
  for (i=0; i<MAXCLIPS; i++)  clippause[i]  = -1;

  // X Variable Declarations
  Display  *D;
  Visual   *V;
  Colormap  C;
  int       S,Loop,Point;
  Window    W;
  GC        Gc, GcRev;
  XImage   *I;
  XColor    Colors[256];
  XSetWindowAttributes Attr;
  XVisualInfo Info;
  char     *Data;
  int       GlobalWidth  = 250;
  int       GlobalHeight = 100;
  XEvent    event;
  char      keychar[1];
  KeySym    test_keysym;
  int       buttonheld1 = -1;
  int       buttonheld2 = -1;
  int       buttonheld3 = -1;

  int       sidelength = (float (vw)) * (0.140625);

  // Some of the vision interface
  float maxcolor;
  if (strlen(modelname)>1)
    {
      ColorMixture colorExpert;
      colorExpert.loadMixtureModel(modelname);
      maxcolor = makeProb(&colorExpert, &colorLUT);
      threshold = threshold*maxcolor;
    }
  else
    {
      fprintf(stderr,"Please specify a color model name.\n");
      exit(0);
    }
  if (clip<1.0)
    {
      fprintf(stderr,"Thresholding...\n");
      for (i=0; i<(64*64*64); i++)
	{
	  if (colorLUT[i] < (clip*maxcolor))	colorLUT[i] = 0.0;
	  else colorLUT[i] = maxcolor;
	}
      threshold = maxcolor*0.5;
    }
  TCharImage connImage(v->width()/ss, v->height()/ss);
  float threshold  =    0.1;




  if (keyboard<1)
    {
      // X Setup
      D = XOpenDisplay("");
      S = DefaultScreen(D);
      if (XMatchVisualInfo(D,S,8,PseudoColor,&Info)==0)
	{
	  printf("Display can not handle %d bit pseudo color\n", 8);
	  return;
	}
      V = Info.visual;
      for (i=0; i<256; i++)
	{
	  Colors[i].pixel = i;
	  Colors[i].red   = 0; Colors[i].green = 0; Colors[i].blue = 0;
	  Colors[i].flags = DoRed | DoGreen | DoBlue;
	}
      C = XCreateColormap(D,RootWindow(D,S),V,AllocAll);
      XStoreColors(D,C,Colors,256);
      Attr.colormap = C;
      Attr.background_pixel = WhitePixel(D,S);
      Attr.border_pixel = WhitePixel(D,S);
      W = XCreateWindow(D,DefaultRootWindow(D),0,0,GlobalWidth,
			GlobalHeight,100,8,InputOutput,V,
			CWColormap|CWBackPixel|CWBorderPixel,&Attr);
      XStoreName(D,W,"DyPERS - Jebara, Oliver, Schiele");
      XMapRaised(D,W);
      Gc = XCreateGC(D,W,0L, (XGCValues *) 0);
      Data = (char *) malloc(GlobalWidth*GlobalHeight*sizeof(char));
      for (i=0; i<(GlobalWidth*GlobalHeight); i++) Data[i]=0;
      I = XCreateImage(D,V,8,ZPixmap,0,Data,GlobalWidth,GlobalHeight,8,0);
      XSelectInput (D, W, KeyPressMask | ButtonPressMask | ButtonReleaseMask | StructureNotifyMask);
    }

  // Audio Setup
  mic.SetQSize(MAXAUDIO);
  mic.Open();

  if (keyboard==1)
    {
      // Setup the Keyboard
      ioctl(STDIN_FILENO, TCGETS, &OrigTerm);
      ioctl(STDIN_FILENO, TCGETS, &TermDescr);
      TermMask=04401;
      TermDescr.c_lflag=TermDescr.c_lflag & TermMask;
      TermDescr.c_cc[VEOF]=1;
      TermDescr.c_cc[VEOL]=0;
      ioctl(STDIN_FILENO, TCSETS, &TermDescr);
    }

  // Main Loop
  while (1)
    {
      // Grab an Image and Show it with the Objective Circle
      current = v->snap();
      v->blit(current);
      
      // Check if black or not black
      illum = 0.0;
      for (j=0; j<v->height; j+=10)
	for (i=0; i<v->width; i+=10)
	  {
	    illum += (((current[j*v->width+i] & 0x0000FF))  +
		      ((current[j*v->width+i] & 0x00FF00) >> 8)+
		      ((current[j*v->width+i] & 0xFF0000) >> 16));
	  }
      if (illum<bright) // blah blah



      if (buttonheld1>0)
	{
	  v->box(vw/2-sidelength, vh/2-sidelength,vw/2+sidelength, vh/2+sidelength,-2);
	  v->box(vw/2-sidelength+1, vh/2-sidelength+1,vw/2+sidelength+1, vh/2+sidelength+1,-2);
	  v->box(vw/2-sidelength+2, vh/2-sidelength+2,vw/2+sidelength+2, vh/2+sidelength+2,-2);
	  v->box(vw/2-sidelength+3, vh/2-sidelength+3,vw/2+sidelength+3, vh/2+sidelength+3,-2);

	}
      else if (buttonheld2>0)
	{
	  v->box(vw/2-sidelength, vh/2-sidelength,vw/2+sidelength, vh/2+sidelength,-3);
	  v->box(vw/2-sidelength+1, vh/2-sidelength+1,vw/2+sidelength+1, vh/2+sidelength+1,-3);
	  v->box(vw/2-sidelength+2, vh/2-sidelength+2,vw/2+sidelength+2, vh/2+sidelength+2,-3);
	  v->box(vw/2-sidelength+3, vh/2-sidelength+3,vw/2+sidelength+3, vh/2+sidelength+3,-3);
	}
      else if (buttonheld3<0)
	{
	  v->box(vw/2-sidelength, vh/2-sidelength,vw/2+sidelength, vh/2+sidelength,-4);
	  v->box(vw/2-sidelength+1, vh/2-sidelength+1,vw/2+sidelength+1, vh/2+sidelength+1,-4);
	  v->box(vw/2-sidelength+2, vh/2-sidelength+2,vw/2+sidelength+2, vh/2+sidelength+2,-4);
	  v->box(vw/2-sidelength+3, vh/2-sidelength+3,vw/2+sidelength+3, vh/2+sidelength+3,-4);
	}


      if (buttonheld3>0)
	{
	  // Record while the user holds the button
	  totalframes += mic.Read();
	  record();
	}
      else
	{
	  // Clear the X Window
	  if (keyboard<1) { XPutImage(D,W,Gc,I,0,0,0,0,GlobalWidth,GlobalHeight); XFlush(D); }
	  // Otherwise take a break and
	  // check for object recognition and playback if needed
	  usleep(10000);
	  if (server!=NULL)
	    {
	      cldypers.getRecObject(recRes);
	      if (recRes.id!=-2) objectlabel = recRes.id;
	    }
	  // Playback the Current Recording
	  if ((objectlabel>=0) && (objectlabel<MAXLABELS))
	    {
	      playback(objectlabel);
	    }
	  else playback(-1);
	}

      if (keyboard==1)
	{
	  // Grab from Keyboard
	  ioctl(0,FIONREAD,&Key);
	  if (Key>0)
	    {
	      Key=getchar();
	      if (Key=='.')
		{
		  cldypers.Dispose();
		  printf("Quitting\n");
		  exit(0);

		}
	      if (Key=='-')
		{
		  // Recording ..
		  if (buttonheld3<0)
		    {
		      printf("Recording\n");
		      // On Right Mouse Down, increment recClip
		      // Keep grabbing while right mouse down
		      if (recClip<0)
			{
			  recClip = 0;
			  clipstart[recClip] = recPic;
			  record();
			}
		      else if (recClip<MAXCLIPS)
			{
			  if (buttonheld3 < 0)
			    {
			      recClip++;
			      clipstart[recClip] = recPic;
			      record();
			    }
			  else
			    {
			      record();
			    }
			}
		      totalframes = 0;
		      buttonheld3 = 1;
		    }
		  else
		    {
		      printf("Done Recording\n");
		      // On Right Mouse Up, stop grabbing
		      buttonheld3         = -1;
		      if (totalframes>=MAXAUDIO)
			totalframes = MAXAUDIO;
		      audioptr            = mic.GetLast(totalframes);
		      audioclips[recClip] = (float *) malloc(totalframes*sizeof(float));
		      audiolen[recClip]   = totalframes;
		      audioclip           = audioclips[recClip];
		      for (i=0; i<totalframes; i++) audioclip[i] = *(audioptr++);
		    }
		}
	      if (Key=='+')
		{
		  // Associating
		  buttonheld1 = 1;
		  buttonheld1 = -1;
		  associate();
		  printf("Associated\n");
		}
	      if (Key=='9')
		{
		  // Garbage
		  buttonheld2 = 1;
		  buttonheld2 = -1;
		  garbage();
		  printf("Garbage\n");
		}
	    }
	}
      else
	{
	  // Process User Events
	  if (XCheckMaskEvent(D, KeyPressMask | ButtonPressMask | ButtonReleaseMask | StructureNotifyMask, &event))
	    {
	      switch (event.type)
		{ 
		case ButtonPress:
		  if (event.xbutton.button == Button1)
		    {
		      buttonheld1 = 1;
		    }
		  if (event.xbutton.button == Button2)
		    {
		      buttonheld2 = 1;
		    }
		  if (event.xbutton.button == Button3)
		    {
		      // On Right Mouse Down, increment recClip
		      // Keep grabbing while right mouse down
		      if (recClip<0)
			{
			  recClip = 0;
			  clipstart[recClip] = recPic;
			  record();
			}
		      else if (recClip<MAXCLIPS)
			{
			  if (buttonheld3 < 0)
			    {
			      recClip++;
			      clipstart[recClip] = recPic;
			      record();
			    }
			  else
			    {
			      record();
			    }
			}
		      totalframes = 0;
		      buttonheld3 = 1;
		    }
		  break;
		case ButtonRelease:
		  if (event.xbutton.button == Button1)
		    {
		      // On Left Mouse Click, connect recClip to current label and increment label
		      buttonheld1 = -1;
		      associate();
		    }
		  if (event.xbutton.button == Button2)
		    {
		      buttonheld2 = -1;
		      garbage();
		    }
		  if (event.xbutton.button == Button3)
		    {
		      // On Right Mouse Up, stop grabbing
		      buttonheld3         = -1;
		      if (totalframes>=MAXAUDIO)
			totalframes = MAXAUDIO;
		      audioptr            = mic.GetLast(totalframes);
		      audioclips[recClip] = (float *) malloc(totalframes*sizeof(float));
		      audiolen[recClip]   = totalframes;
		      audioclip           = audioclips[recClip];
		      for (i=0; i<totalframes; i++) audioclip[i] = *(audioptr++);
		    }
		  break;
		case KeyPress:
		  {
		    XLookupString(&(event.xkey), keychar, 1, &test_keysym, NULL);
		    if (!(IsCursorKey(test_keysym) || IsFunctionKey(test_keysym) ||
			  IsKeypadKey(test_keysym) || IsMiscFunctionKey(test_keysym) ||
			  IsModifierKey(test_keysym) || IsPFKey(test_keysym) ||
			  test_keysym ==  XK_Multi_key ))
		      {
			switch (keychar[0])
			  {
			  case 'q':
			    // Free up the ram!
			    mic.Close();
			    if (server!=NULL) cldypers.Dispose();
			    exit(0); 
			    break;
			  case 'a':
			    objectlabel = -1; break;
			  case 'b':
			    objectlabel = 0; break;
			  case 'c':
			    objectlabel = 1; break;
			  case 'd':
			    objectlabel = 2; break;
			  case 'e':
			    objectlabel = 3; break;
			  case 'f':
			    objectlabel = 4; break;
			  }
		      }
		    break;
		  }
		default:
		  break;
		}
	    }
	}
    }
}


